# Tools for profiling.
import inspect
import gc
import os
import sys
import time
from functools import wraps
from collections import Counter


# External module.
try:
    import psutil
except ImportError:
    # Create a custom local version of 'psutil'. It will be slower, but it will work.
    SHOW_PSUTIL_WARNING = True
    class psutil:
        pid = 0
        rss_for_pid = "ps -p {pid} -o rss="
        def __init__(self):
            # Generate a warning about using a slower method for retrieving memory usage.
            global SHOW_PSUTIL_WARNING
            if (SHOW_PSUTIL_WARNING):
                import logging
                logging.warning("Using custom 'psutil' class that relies on calls to 'os.system'. It will be much slower than if you install 'psutil' with:\n  python3 -m pip install --user psutil\n")
                SHOW_PSUTIL_WARNING = False
            # Set the PID.
            import os
            self.pid = os.getpid()
            # Get the platform, if it Windows then update the command for getting memory utilization.
            import platform
            if platform.system() == 'Windows':
                self.rss_for_pid = "wmic process where ProcessID={pid} get WorkingSetSize"
        def Process():
            return psutil()
        def memory_info(self):
            return self
        @property
        def rss(self):
            # Execute the system command that gets the resident memory size of a process for this platform.
            mem_use_kb = os.popen(self.rss_for_pid.format(pid=self.pid)).read()
            # Return the memory usage in bytes as an integer.
            return int(mem_use_kb) * 1024


# Declare the divisors for different memory units.
MEM_UNIT_DIVISORS = {
    "KB" : 2**10,
    "MB" : 2**20,
    "GB" : 2**30,
    "TB" : 2**40,
}


# Return the memory usage currently of this process.
def mem_use(proc=psutil.Process(), unit="MB"):
    return proc.memory_info().rss / MEM_UNIT_DIVISORS[unit]


# Inspect the count and highest memory consuming objects.
def inspect_memory(top=10, max_history=1, history=[]):
    # Count the number of objects in memory and the total size.
    c = Counter(type(o).__name__ for o in gc.get_objects())
    s = sum(sys.getsizeof(o) for o in gc.get_objects())
    size_str   = f":: Total size: {s / (2**20):.2f}MB"
    top_str    = f":: Top obj counts: {c.most_common(top)}"
    # Measure the change in counts.
    if (len(history) > 0):
        d = sorted(((k, c.get(k,0) - history[-1].get(k,0)) for k in set(c)|set(history[-1])),
                   key = lambda i: (-i[1], i[0]))[:top]
        change_str = f":: Top count changes: {d}"
    else:
        change_str = ""
    width = max(len(size_str), len(top_str), len(change_str))
    # Print the update.
    print(":"*width)
    print(size_str)
    print(top_str)
    if (len(change_str) > 0): print(change_str)
    print(":"*width)
    # Update the history.
    history.append(c)
    while (len(history) > max_history):
        history.pop(0)


# Generate a summary string for a profile of a function.
def summarize_profile(profile, munit="MB", tunit="s", mdiv=2**20, tdiv=1):
    if len(profile) == 0:
        return "The function has not yet been executed."
    lines = []
    # Column widths.
    widths = [
        len("Time"),
        len("MDelta"),
        len("Mem"),
        len("Execs"),
        0,  # length of the code strings
    ]
    # Get the file ane function names.
    file_name, function_name, _, _ = next(iter(profile))
    # Get the totals for memory and time changes.
    true_time = 0
    time_total = 0
    mem_total = 0
    pos_mem_total = 0
    for (_, _, line, code), (times, mdelta, mem, execs) in profile.items():
        true_time += sum(times)
        if line != "call":
            time_total += sum(times)
            mem_total += sum(mdelta)
            pos_mem_total += sum((v for v in mdelta if v > 0))
    # Iterate over the lines of code and their stats.
    for (_, _, line, code), (times, mdelta, mem, execs) in sorted(
            profile.items(), key=lambda k: 0 if k[0][2] == "call" else int(k[0][2])
    ):
        # Collapse the lists to specific values.
        time = sum(times)
        mdelta = sum(mdelta)
        mem = (min(mem), max(mem))
        # Store the profiling details for this line.
        mem = f"[{mem[0]/mdiv:.2f}{munit}, {mem[1]/mdiv:.2f}{munit}]"
        if line == "call":
            time = ""
            mdelta = ""
        else:
            time = f"{time / tdiv:.2f}{tunit} ({100.0 * time / time_total if time_total > 0 else 0:.0f}%)"
            mdelta = f"{mdelta / mdiv:.2f}{munit} ({100.0 * mdelta / pos_mem_total if pos_mem_total > 0 else 0:.0f}%)"
        execs = f"{execs}"
        # Store the max of the column widths.
        widths[:] = (max(ov, len(nv)) for (ov, nv) in zip(
            widths, (time, mdelta, mem, execs, code)
        ))
        # Append the output to include in the string summary.
        lines.append((line, time, mdelta, mem, execs, code))
    # Generate the summary string.
    string = f"Line   {'Time':<{widths[0]}s}  {'MDelta':<{widths[1]}s}  {'Mem':<{widths[2]}s}  {'Execs':<{widths[3]}s}  Code"
    bar = "-" * (len(string) + 4) + "\n"
    string = (
        "_" * (len(string) + 4)
        + "\n"
        + f"File:     {file_name}\n"
        + f"Function: {function_name}\n"
        + f"Memory:   {mem_total / mdiv:.2f}{munit}\n"
        + f"Time:     {true_time / tdiv:.2f}{tunit} ({time_total / tdiv:.2f}{tunit} accounted for)\n"
        + string + "\n"
        + bar
    )
    for (number, time, mdelta, mem, execs, code) in lines:
        string += f"{number:5s}  {time:{widths[0]}s}  {mdelta:{widths[1]}s}  {mem:{widths[2]}s}  {execs:{widths[3]}s}  {code:<{widths[4]}s}\n"
    # Add one last line.
    string += bar
    return string


# A decorator that can be used to show the line-by-line memory and time consumtion of a python function.
def profile(f):
    # Store some local variables that can be reused in the scope of the profiling decorator.
    process = psutil.Process()
    profile = {}
    previous = dict(
        line = "call",
        code = "",
        time = time.time(),
        mem = 0,
        depth = 0,
    )
    # Create a function that monitors memory usage and elapsed time between executed lines of python code.
    # Arguments include:
    #   frame - the frame object (from insepct or sys.trace);
    #   event - one of 'call', 'exception', 'line', 'opcode', or 'return';
    #   arg - something that we do not use here.
    def track(frame, event, arg):
        nonlocal process, profile, previous
        # Depth increments whenever a call is made and we can use this to block tracking in child functions.
        if event == "call": previous['depth'] += 1
        # When we encounter a line that is in the target function, we record stats.
        if (previous['depth'] == 1) and (event in {"call", "line", "return"}):
            # Store the time delta (since last line).
            elapsed = time.time() - previous["time"]
            # Get information about the current scope.
            frame_info = inspect.getframeinfo(frame)
            file_name = os.path.join(
                os.path.basename(os.path.split(frame_info.filename)[0]),
                os.path.basename(frame_info.filename)
            )
            function_name = frame_info.function
            line, code = previous["line"], previous["code"]
            # Create a hashable identifier for this line.
            line_id = (file_name, function_name, line, code)
            # Retreive / create profiling information for this line.
            profile_vals = profile.get(line_id, [[], [], [], 0])  # [time, mem delta, mem, executions]
            # Get and update histories.
            mem = process.memory_info().rss
            profile_vals[0].append(elapsed)
            profile_vals[1].append(mem - previous["mem"])
            profile_vals[2].append(mem)
            profile_vals[3] += 1
            # Store the updated history for this line.
            profile[line_id] = profile_vals
            # Update the "code" and "line number" for the next iteration.
            previous["code"] = ("".join(frame_info.code_context)).rstrip().replace("\n", "; ")
            previous["line"] = str(frame.f_lineno)
            previous["mem"] = mem
            previous["time"] = time.time()
        # When a return event occurs, decrement the depths. Also empty history each time the function fully exits.
        if event == "return":
            previous["depth"] -= 1
            if previous["depth"] == 0:
                previous["code"] = ""
                previous["line"] = "call"
        # Trace fuctions return the "next" trace function, in this case, itself (to be called again later).
        return track
    # 
    # Now we create a function that wraps the original function with added tracking.
    @wraps(f)
    def profiled(*args, **kwargs):
        nonlocal f
        try:
            # Get any existing trace (in case multiple python debuggers are executing).
            existing_trace = sys.gettrace()
            if existing_trace is None:
                # Assign the memory tracking function as a trace function (to be called by the Python interpreter).
                sys.settrace(track)
            else:
                # Generate a new trace function that executes this memory trace and the existing one.
                def stacked_trace(*args, **kwargs):
                    nonlocal track, existing_trace
                    track = track(*args, **kwargs)
                    existing_trace = existing_trace(*args, **kwargs)
                    return stacked_trace

                # Set the trace to the new stacked tracing function.
                sys.settrace(stacked_trace)
            # Execute the original (now decorated) function.
            result = f(*args, **kwargs)
        # Reset the trace to its state before this decorator.
        finally:
            sys.settrace(existing_trace)
        return result
    # 
    profiled.nonlocals = {
        "function": f,
        "process": process,
        "profile": profile,
        "previous": previous,
    }
    profiled.profile_summary = lambda: summarize_profile(profile)
    profiled.show_profile = lambda: print(summarize_profile(profile))
    # 
    return profiled




import numpy as np
from tlux.random import well_spaced_box
from tlux.approximate.axy import AXY
from tlux.plot import Plot

# A function for testing approximation algorithms.
def f(x):
    x = x.reshape((-1,2))
    x, y = x[:,0], x[:,1]
    return (3*x + np.cos(8*x)/2 + np.sin(5*y))



x = well_spaced_box(100, 2)
y = f(x)

print("x: ", x.shape)
print("y.shape: ", y.shape)

m = AXY()
m.fit(x=x, y=y)

p = Plot()
p.add("data", *x.T, y)
p.add_func("mode", m, *([[-.1, 1.1]]*2), vectorized=True)
p.show()



import numpy as np
import fmodpy

# Matrix operations module.
mops = fmodpy.fimport("../axy_matrix_operations.f90", blas=True, lapack=True).matrix_operations
print()


# --------------------------------------------------------------------
#                         ORTHOGONALIZE

# Tested up to ~400 dimension, where it fails witih precision=5.
def _test_orthogonalize(max_dimension=64, trials=100, precision=5):
    print("orthogonalize..", end=" ", flush=True)
    # Test orthogonalization.
    np.set_printoptions(linewidth=1000)
    for dimension in range(1, max_dimension+1):
        for trial in range(trials):
            np.random.seed(trial)
            matrix = np.random.random(size=(dimension,dimension)).astype("float32")
            lengths = np.zeros(dimension, dtype="float32")
            ortho = matrix.copy()
            mops.orthogonalize(ortho.T, lengths)
            if (min(lengths) == 0.0): continue
            identity = np.identity(dimension).astype("float32")
            AtA = ortho.T @ ortho
            assert (abs(AtA-identity).max()) < 10**(-precision), \
                "Results of (At A) are not orthogonal enough.\n" \
                f"  Dimension = {dimension}\n" \
                f"  Trial = {trial}\n\n" \
                f"{(AtA).round(precision)}\n\n" \
                f"matrix = np.asarray({matrix.tolist()}, dtype='float32')\n" \
                f"lengths = np.zeros({dimension}, dtype='float32')\n" \
                "ortho = matrix.copy()\n" \
                f"mops.orthogonalize(ortho.T, lengths)\n" \
                f"bad_result = ortho.T @ ortho"
            AAt = ortho @ ortho.T
            assert (abs(AAt-identity).max()) < 10**(-precision), \
                "Results of (A At) are not orthogonal enough.\n" \
                f"  Dimension = {dimension}\n" \
                f"  Trial = {trial}\n\n" \
                f"{(AAt).round(precision)}\n\n" \
                f"matrix = np.asarray({matrix.tolist()}, dtype='float32')\n" \
                f"lengths = np.zeros({dimension}, dtype='float32')\n" \
                "ortho = matrix.copy()\n" \
                f"mops.orthogonalize(ortho.T, lengths)\n" \
                f"bad_result = ortho @ ortho.T"
    print("passed.", flush=True)

_test_orthogonalize()


# --------------------------------------------------------------------
#                         RADIALIZE

# Generate a random rotation matrix (that rotates a random amount along each axis).
def random_rotation(dimension):
    # Determine the order of the dimension rotations (done in pairs)
    rotation_order = np.arange(dimension)
    # Generate the rotation matrix by rotating two dimensions at a time.
    rotation_matrix = np.identity(dimension)
    for (d1, d2) in zip(rotation_order, np.roll(rotation_order,1)):
        next_rotation = np.identity(dimension)
        # Establish the rotation
        rotation = np.random.random() * 2 * np.pi
        next_rotation[d1,d1] =  np.cos(rotation)
        next_rotation[d2,d2] =  np.cos(rotation)
        next_rotation[d1,d2] =  np.sin(rotation)
        next_rotation[d2,d1] = -np.sin(rotation)
        # Compound the paired rotations
        rotation_matrix = np.matmul(next_rotation, rotation_matrix)
        # When there are two dimenions or fewer, do not keep iterating.
        if (dimension <= 2): break
    return rotation_matrix

# Generate random data with skewed distribution along the principle axes.
def random_data(num_points, dimension, box=10, skew=lambda x: 1 * x**2 / sum(x**2)):
    center = (np.random.random(size=(dimension,)) * box - box/2).astype("float32")
    variance = skew(np.random.random(size=(dimension,))).astype("float32")
    data = np.random.normal(center, variance, size=(num_points,dimension)).astype("float32")
    rotation = random_rotation(dimension).astype("float32")
    return np.asarray(data @ rotation), center, variance, rotation

# Test function for visually checking the "radialize" function.
def _test_radialize(show=True):
    print("radialize..", end=" ", flush=True)
    # Generate data to test with.
    num_trials = 10
    np.random.seed(0)
    num_points = 1000
    dimension = 3
    from tlux.plot import Plot
    p = Plot()
    for i in range(num_trials):
        # Generate random data (that is skewed and placed off center).
        x, shift, scale, rotation = random_data(num_points, dimension)
        p.add(str(i+1), *x.T, marker_size=3)

        # Generate the radialized version.
        shift = np.zeros(dimension, dtype="float32")
        transform = np.zeros((dimension, dimension), dtype="float32", order="F")
        inverse = np.zeros((dimension, dimension), dtype="float32", order="F")
        to_flatten = (i % 2) == 0
        descriptor = 'radialized' if to_flatten else 'normalized'
        xr, shift, transform, inverse = mops.radialize(
            x=(x.copy()).T, shift=shift, vecs=transform, inverse=inverse,
            max_to_flatten=(0 if to_flatten else None),
        )
        p.add(f"{i+1} {descriptor}", *xr, marker_size=3)

        # Use the inverse to "fix" the data back to its original form.
        xf = (inverse.T @ xr).T
        xf -= shift
        p.add(f"{i+1} fixed", *xf.T, marker_size=3)

        # Use the provided shift and transform to repeate the radialization process.
        xrr = (x + shift) @ transform
        p.add(f"{i+1} re {descriptor}", *xrr.T, marker_size=3)
    if (show): p.show()
    print("passed.", flush=True)

_test_radialize(show=False)


# --------------------------------------------------------------------
#                            LEAST_SQUARES

def _test_least_squares():
    # Use the matrix operations subroutine to solve the least squares problem.
    br = 2
    d = 3
    mo_x = np.zeros((d,br), order="F", dtype="float32")
    mops.least_squares(
        bytes("T","ascii"),
        np.array(a.T, order="F", dtype="float32"),
        np.array(b, order="F", dtype="float32"),
        mo_x)

    np_x, residual = np.linalg.lstsq(a, b, rcond=None)[:2]

    print()
    print("x: ")
    print(x)

    print()
    print("np_x: ")
    print(np_x)

    print()
    print("mo_x: ")
    print(mo_x)

    print()
    print("mo_x (scaled to full output):")
    upscale = np.linalg.lstsq(np.matmul(a, mo_x), b, rcond=None)[0]
    print(np.matmul(mo_x, upscale))
    print()


    # Generate a visiaul.
    from tlux.plot import Plot
    p = Plot()
    p.add("a", *a.T, color=1)
    # Show the true values.
    for i in range(bd):
        p.add(f"b{i}", *a.T, b[:,i], color=2+i, shade=True)
    # Show the literal values produced by the mo regression.
    for i in range(br):
        p.add(f"app b{i}", *a.T, np.matmul(a, mo_x)[:,i],
              color=(0,0,0,0.8), marker_size=2)
    # Show the original approximations recovered from the mo regression.
    for i in range(bd):
        p.add(f"recon b{i}", *a.T, np.matmul(a, np.matmul(mo_x, upscale))[:,i],
              color=2+i, shade=True, marker_size=4, marker_line_width=1)

    # Show the plot, different depending on if it is 2D (points only) or 3D (with regression).
    if (bd > 0):
        p.show(x_range=[-1.1,1.1], y_range=[-1.1,1.1], z_range=[-1.1,1.1])
    else:
        p.show(x_range=[-2.3,2.3], y_range=[-1.1,1.1])
    print()


# _test_least_squares()



from tlux.data.exceptions import BadValue, ImproperUsage

# This class is a protected type of list where the contents are
# supposed to be of a specific type. It enforces that assignment and
# append operations do not violate the type specifiation of the object.
class Descriptor:
    # Initialize this descriptor, make sure all elements of 'values'
    # have the same type if provided.
    def __init__(self, data, values=None, types=type(None)):
        # Store the parent data object of this descriptor.
        self.data = data
        # Store the provided data (or make a new set)
        if (values is None): values = []
        self.values = values
        # Initialize the type.
        self.type = types
        for value in self.values:
            if (self.type == type(None)): self.type = type(value)
            elif (self.type != type(value)):
                raise(BadValue(f"Only type '{self.type}' objects can be added to this descriptor."))

    # Check for the type of the added value.
    def append(self, value):
        if (self.data.view): raise(ImproperUsage("Cannot modify descriptor of view."))
        if (self.type == type(None)): self.type = type(value)
        if (self.type != type(value)):
            raise(BadValue(f"Only type '{self.type}' objects can be added to this descriptor."))
        return self.values.append(value)

    # Check for type in the "setitem" operation.
    def __setitem__(self, index, value):
        if (self.data.view): raise(ImproperUsage("Cannot modify descriptor of view."))
        if (self.type == type(None)): self.type = type(value)
        if (self.type != type(value)):
            raise(BadValue(f"Only type '{self.type}' objects can be added to this descriptor."))
        self.values[index] = value

    # Return the appropriately mapped values through getitem.
    def __getitem__(self, index):
        # If this is a slice, then return a 
        if (type(index) == slice):
            return [self.values[self.data.col(i)] for i in range(self.data.shape[1])[index]]
        # Use the standard getitem operator.
        return self.values[self.data.col(index)]

    # Define a "len" operator.
    def __len__(self): return self.data.shape[1]

    # Define a "str" operator.
    def __str__(self): return str(list(self))

    # Define an "insert" function.
    def insert(self, idx, value):
        if (self.data.view): raise(ImproperUsage("Cannot modify descriptor of view."))
        if (self.type == type(None)): self.type = type(value)
        if (self.type != type(value)):
            raise(BadValue(f"Only type '{self.type}' objects can be added to this descriptor."))
        # Add the new "value" at a desired "index".
        return self.values.insert(idx, value)

    # Identify the index (column number) of a value.
    def index(self, value): return self.values.index(value)
    # TODO: ^^ Modify this code if the data is a view.
    #       return self.data._col_indices.index(self.values.index(value))

    # Pop a value out of this descriptor.
    def pop(self, index):
        if (self.data.view): raise(ImproperUsage("Cannot modify descriptor of view."))
        return self.values.pop(index)
        

# Local mutable Column class (for column item assignment,
# retrieval, and iteration over a column in a data object).
class Column:
    def __init__(self, data, column, indices=None):
        # Generate indices if they were not provided
        if (indices is None): indices = list(range(len(data)))
        # Store the parent data and the column number 
        self.data = data
        self.column = column
        self.indices = indices

    #     Indexing
    # -----------------
    def __setitem__(self, index, value):
        # Default setitem for the parent data
        self.data[self.indices[index], self.column] = value
    def __getitem__(self, index):
        # Default getitem for the parent data, bypass "getitem" and
        # retrieve value stored in internal data.
        if (type(index) == int):
            return self.data.data[self.data.row(self.indices[index])][self.data.col(self.column)]
        elif (type(index) == slice):
            indices = range(len(self.indices))[index]
        elif (hasattr(index, "__iter__")):
            # Convert provided index into a list of integers.
            indices = []
            for i,v in enumerate(index):
                if (type(v) == bool):
                    if v: indices.append(self.indices[i])
                elif (type(v) == int):      indices.append(self.indices[v])
                else: raise(self.data.BadIndex(f"Data column index iterable contained type {type(v)}, expected {int} or {bool}."))
        else:
            raise(self.data.BadIndex(f"Data column does not support indexing with type {type(index)}."))
        # Return a new column with modified indices.
        return self.data.Column(self.data, self.column, indices)

    #     Iterator
    # ----------------
    # Define this as an interator
    def __iter__(self):
        class ColIterator:
            i = 0
            col = self
            def __next__(self):
                if (self.i >= len(self.col.indices)):
                    self.i = 0
                    raise(StopIteration)
                self.i += 1
                return self.col.data[self.col.indices[self.i-1], self.col.column]
        return ColIterator()

    #    Descriptors
    # -----------------
    # Get the length of this Column
    def __len__(self): return len(self.indices)
    # Use the iteration to generate a string of this column
    def __str__(self): return str(list(self))

    #     Operators
    # -----------------
    # Generic boolean operator function.
    def _gen_operator(self, other, op_name):
        # First check for length.
        if hasattr(other, "__len__"):
            # If the "other" is not of the contained type.
            if (type(other) != self.data.types[self.column]):
                # If the length doesn't match, raise error.
                if (len(other) != len(self)):
                    raise(self.data.BadValue(f"Length '{len(other)}' does not have the same number of elements as this Column, {len(self)}."))
                # Return pairwise comparison for equal-length objects.
                for (v1, v2) in zip(self, other):
                    if ((v1 is None) or (v2 is None)): yield None
                    else:                              yield getattr(v1, op_name)(v2)
            else:
                # Otherwise return singleton comparison.
                for v in self:
                    if (v is None): yield None
                    else:           yield getattr(v, op_name)(other)
        # Without a length, it must be a singleton comparison.
        else:
            for v in self:
                if (v is None): yield None
                else:           yield getattr(v, op_name)(other)

    # Custom equality operator.
    def _custom_equality(self, other, equality=True):
        # Try to convert given group to an iterable type for comparison.
        if ((type(other) == self.data.types[self.column]) or (other is None)):
            # Return indices where single value equals column values.
            for v in self:
                if (equality and (v == other)):       yield True
                elif (not equality) and (v != other): yield True
                else:                                 yield False
        elif (type(other) == type(self)):
            # Return indices where the two columns are equal (or not).
            for (v1, v2) in zip(self, other):
                if (equality and (v1 == v2)):       yield True
                elif (not equality) and (v1 != v2): yield True
                else:                               yield False
        elif (type(other) == set):
            # Iterate over rows, generating indices that match the equality.
            for v in self:
                if (equality and (v in other)):             yield True
                elif ((not equality) and (v not in other)): yield True
                else:                                       yield False
        elif (hasattr(other, "__iter__")):
            # Iterate over both this column and the other, checking elementwise equality.
            for (i, (v1, v2)) in enumerate(zip(self, other)):
                yield v1 == v2
            if (i != len(self)):
                raise(RuntimeError(f"The other object of type {type(other)} provided an iterator, but had length {i} and does not match expected length {len(self)}. "))
        else:
            raise(self.data.BadData(f"There is no defined equality operator for the provided type {type(other)}, when it does not match expected type {self.data.types[self.column]}."))

    # Define the custom inequality operator using custom equality.
    def _custom_inequality(self, other): return self._custom_equality(other, equality=False)

    # Define the "in" method to return True for contained values.
    def __in__(self, value): return any(v == value for v in self)
    # Equality operator.
    def __eq__(self, other): return self._custom_equality(other)
    # Inequality operator.
    def __ne__(self, other): return self._custom_inequality(other)
    # Less than operator.
    def __lt__(self, other): return self._gen_operator(other, "__lt__")
    # Greater than operator.
    def __gt__(self, other): return self._gen_operator(other, "__gt__")
    # Less than equal operator.
    def __le__(self, other): return self._gen_operator(other, "__le__")
    # Greater than equal operator.
    def __ge__(self, other): return self._gen_operator(other, "__ge__")
    # Addition operator.
    def __add__(self, other): return self._gen_operator(other, "__add__")
    # Subtraction operator.
    def __sub__(self, other): return self._gen_operator(other, "__sub__")
    # Right-side addition operator.
    def __radd__(self, other): return self._gen_operator(other, "__radd__")
    # Right-side subtraction operator.
    def __rsub__(self, other): return self._gen_operator(other, "__rsub__")
    # Multiplication operator.
    def __mul__(self, other): return self._gen_operator(other, "__mul__")
    # True division operator ( a / b ).
    def __truediv__(self, other): return self._gen_operator(other, "__truediv__")
    # Floor division operator ( a // b ).
    def __floordiv__(self, other): return self._gen_operator(other, "__truediv__")


# Local class for iterating over 'rows' of this data object,
# prevents each row from being modified directly by users.
class Row:
    def __init__(self, data, values):
        self.data = data
        self.values = values

    # Convert an index (which may be in a view) to a list of numeric
    # indices that correctly map to the true data.
    def _index_to_nums(self, index):
        # Return integer indices.
        if   type(index) == int:   nums = self.data.col(index)
        # Return string indices (based on parent data).
        elif type(index) == str:   nums = self.data.names.index(index)
        # Return slice indices.
        elif type(index) == slice: nums = list(map(self.data.col, range(len(self))[index]))
        # Convert an iterable into a list of numeric indices.
        elif hasattr(index, "__iter__"):
            nums = []
            i = v = None
            for i,v in enumerate(index):
                if   type(v) == int:          nums.append(self.data.col(v))
                elif type(v) == str:          nums.append(self.data.names.index(v))
                elif (type(v) == bool) and v: nums.append(self.data.col(i))
                else: raise(self.data.BadIndex(f"Index '{index}' not understood by {type(self)} object."))
            # Make sure that boolean indices are full length.
            if (type(v) == bool) and (i != len(self)-1):
                raise(self.data.BadValue(f"Length '{i+1}' boolean index does not have the same number of elements as this Row, {len(self)}."))
        else: raise(self.data.BadIndex(f"Index '{index}' not understood by {type(self)} object."))
        return nums
    # Get an item from this row.
    def __getitem__(self, index):
        idx = self._index_to_nums(index)
        if (type(idx) is int): return self.values[idx]
        else:                  return [self.values[i] for i in idx]
    # Set an item in this row.
    def __setitem__(self, index, value):
        idx = self._index_to_nums(index)
        if (type(idx) is not int):
            if (len(idx) > 1):
                raise(self.data.BadAssignment(f"{type(self)} object can only assign one position at a time."))
            else:
                i = idx[0]
        else:
            i = idx
        # Assign the type of the column of data if it is unassigned.
        if (value is None): pass
        elif self.data.types[i] == type(None):
            self.data.types[i] = type(value)
        elif (self.data.types[i] != type(value)):
            raise(self.data.BadValue(f"Index {i} can only accept {self.data.types[i]}, received {type(value)}."))
        # Assign the value if it is allowable.
        self.values[i] = value
    # Define this as an interator
    def __iter__(self):
        class RowIterator:
            i = 0
            row = self
            def __next__(self):
                if (self.i >= len(self.row)):
                    self.i = 0
                    raise(StopIteration)
                self.i += 1
                return self.row[self.i-1]
        return RowIterator()
    # Define a "length" for this row.
    def __len__(self): return self.data.shape[1]
    def __str__(self): return str(list(self))
    # Define a "insert" function.
    def insert(self, i, value):
        if (self.data.view):
            raise(self.data.ImproperUsage(f"This row is only a view and does not support insertion."))
        # Return the insertion of the new value.
        return self.values.insert(i, value)
    # Define a "pop" function.
    def pop(self, i):
        if (self.data.view):
            raise(self.data.ImproperUsage(f"This row is only a view and does not support 'pop'."))
        # Return the popped value.
        return self.values.pop(i)

    #    Operators
    # ---------------
    # Generic boolean operator function.
    def _gen_operator(self, other, op_name):
        # First check for length.
        if hasattr(other, "__len__"):
            # If the length doesn't match, check if it is a singleton.
            if (len(other) != len(self)):
                # Otherwise return singleton comparison.
                for v in self:
                    if (v is None): yield None
                    else:           yield getattr(v, op_name)(other)
            # Return pairwise comparison for equal-length objects.
            else:
                for (v1, v2) in zip(self, other):
                    if (v1 is None) or (v2 is None): yield None
                    else: yield getattr(v1, op_name)(v2)
        # Without a length, it must be a singleton comparison.
        else:
            for v in self:
                if (v is None): yield None
                else:           yield getattr(v, op_name)(other)

    # Inequality operator.
    def __ne__(self, other): return self._gen_operator(other, "__ne__")
    # Equality operator.
    def __eq__(self, other): return self._gen_operator(other, "__eq__")
    # Less than operator.
    def __lt__(self, other): return self._gen_operator(other, "__lt__")
    # Greater than operator.
    def __gt__(self, other): return self._gen_operator(other, "__gt__")
    # Less than equal operator.
    def __le__(self, other): return self._gen_operator(other, "__le__")
    # Greater than equal operator.
    def __ge__(self, other): return self._gen_operator(other, "__ge__")
    # Addition operator.
    def __add__(self, other): return self._gen_operator(other, "__add__")
    # Subtraction operator.
    def __sub__(self, other): return self._gen_operator(other, "__sub__")
    # Right-side addition operator.
    def __radd__(self, other): return self._gen_operator(other, "__radd__")
    # Right-side subtraction operator.
    def __rsub__(self, other): return self._gen_operator(other, "__rsub__")
    # Multiplication operator.
    def __mul__(self, other): return self._gen_operator(other, "__mul__")
    # True division operator ( a / b ).
    def __truediv__(self, other): return self._gen_operator(other, "__truediv__")
    # Floor division operator ( a // b ).
    def __floordiv__(self, other): return self._gen_operator(other, "__truediv__")



'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.


'''

import os
import ctypes
import platform
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "prune." + platform.machine() + ".so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-fPIC', '-shared', '-O3', '-fopenmp']
_ordered_dependencies = ['prune.f90', 'prune_c_wrapper.f90']
_symbol_files = []# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the prerequisite symbols for the compiled code.
for _ in _symbol_files:
    _ = ctypes.CDLL(os.path.join(_this_directory, _), mode=ctypes.RTLD_GLOBAL)
# Try to import the existing object. If that fails, recompile and then try.
try:
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


class prune:
    ''''''

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine INNER
    
    def inner(self, tree_size, levels, indices, current_level=None, starting_index=None, next_index=None):
        '''! Get the indices of all the inner children in this tree (the 50th
! percentile points) up to level 'LEVELS'.'''
        
        # Setting up "tree_size"
        if (type(tree_size) is not ctypes.c_long): tree_size = ctypes.c_long(tree_size)
        
        # Setting up "levels"
        if (type(levels) is not ctypes.c_long): levels = ctypes.c_long(levels)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "current_level"
        current_level_present = ctypes.c_bool(True)
        if (current_level is None):
            current_level_present = ctypes.c_bool(False)
            current_level = ctypes.c_long()
        else:
            current_level = ctypes.c_long(current_level)
        if (type(current_level) is not ctypes.c_long): current_level = ctypes.c_long(current_level)
        
        # Setting up "starting_index"
        starting_index_present = ctypes.c_bool(True)
        if (starting_index is None):
            starting_index_present = ctypes.c_bool(False)
            starting_index = ctypes.c_long()
        else:
            starting_index = ctypes.c_long(starting_index)
        if (type(starting_index) is not ctypes.c_long): starting_index = ctypes.c_long(starting_index)
        
        # Setting up "next_index"
        next_index_present = ctypes.c_bool(True)
        if (next_index is None):
            next_index_present = ctypes.c_bool(False)
            next_index = ctypes.c_long()
        else:
            next_index = ctypes.c_long(next_index)
        if (type(next_index) is not ctypes.c_long): next_index = ctypes.c_long(next_index)
    
        # Call C-accessible Fortran wrapper.
        clib.c_inner(ctypes.byref(tree_size), ctypes.byref(levels), ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(current_level_present), ctypes.byref(current_level), ctypes.byref(starting_index_present), ctypes.byref(starting_index), ctypes.byref(next_index_present), ctypes.byref(next_index))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, (next_index.value if next_index_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine OUTER
    
    def outer(self, tree_size, levels, indices, current_level=None, starting_index=None, next_index=None):
        '''! Get the indices of all the outer children in this tree (extrema
! points) up to level 'LEVELS'.'''
        
        # Setting up "tree_size"
        if (type(tree_size) is not ctypes.c_long): tree_size = ctypes.c_long(tree_size)
        
        # Setting up "levels"
        if (type(levels) is not ctypes.c_long): levels = ctypes.c_long(levels)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "current_level"
        current_level_present = ctypes.c_bool(True)
        if (current_level is None):
            current_level_present = ctypes.c_bool(False)
            current_level = ctypes.c_long()
        else:
            current_level = ctypes.c_long(current_level)
        if (type(current_level) is not ctypes.c_long): current_level = ctypes.c_long(current_level)
        
        # Setting up "starting_index"
        starting_index_present = ctypes.c_bool(True)
        if (starting_index is None):
            starting_index_present = ctypes.c_bool(False)
            starting_index = ctypes.c_long()
        else:
            starting_index = ctypes.c_long(starting_index)
        if (type(starting_index) is not ctypes.c_long): starting_index = ctypes.c_long(starting_index)
        
        # Setting up "next_index"
        next_index_present = ctypes.c_bool(True)
        if (next_index is None):
            next_index_present = ctypes.c_bool(False)
            next_index = ctypes.c_long()
        else:
            next_index = ctypes.c_long(next_index)
        if (type(next_index) is not ctypes.c_long): next_index = ctypes.c_long(next_index)
    
        # Call C-accessible Fortran wrapper.
        clib.c_outer(ctypes.byref(tree_size), ctypes.byref(levels), ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(current_level_present), ctypes.byref(current_level), ctypes.byref(starting_index_present), ctypes.byref(starting_index), ctypes.byref(next_index_present), ctypes.byref(next_index))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, (next_index.value if next_index_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine TOP
    
    def top(self, tree_size, levels, indices, current_level=None, starting_index=None, next_index=None):
        '''! Get the indices of all the top children in this tree up to level 'LEVELS'.'''
        
        # Setting up "tree_size"
        if (type(tree_size) is not ctypes.c_long): tree_size = ctypes.c_long(tree_size)
        
        # Setting up "levels"
        if (type(levels) is not ctypes.c_long): levels = ctypes.c_long(levels)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "current_level"
        current_level_present = ctypes.c_bool(True)
        if (current_level is None):
            current_level_present = ctypes.c_bool(False)
            current_level = ctypes.c_long()
        else:
            current_level = ctypes.c_long(current_level)
        if (type(current_level) is not ctypes.c_long): current_level = ctypes.c_long(current_level)
        
        # Setting up "starting_index"
        starting_index_present = ctypes.c_bool(True)
        if (starting_index is None):
            starting_index_present = ctypes.c_bool(False)
            starting_index = ctypes.c_long()
        else:
            starting_index = ctypes.c_long(starting_index)
        if (type(starting_index) is not ctypes.c_long): starting_index = ctypes.c_long(starting_index)
        
        # Setting up "next_index"
        next_index_present = ctypes.c_bool(True)
        if (next_index is None):
            next_index_present = ctypes.c_bool(False)
            next_index = ctypes.c_long()
        else:
            next_index = ctypes.c_long(next_index)
        if (type(next_index) is not ctypes.c_long): next_index = ctypes.c_long(next_index)
    
        # Call C-accessible Fortran wrapper.
        clib.c_top(ctypes.byref(tree_size), ctypes.byref(levels), ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(current_level_present), ctypes.byref(current_level), ctypes.byref(starting_index_present), ctypes.byref(starting_index), ctypes.byref(next_index_present), ctypes.byref(next_index))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, (next_index.value if next_index_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine LEVEL
    
    def level(self, tree_size, layer, indices, found, starting_index=None):
        '''! Get the indices of the level "LAYER" of this tree, with 0 being
! the root. List of indices is returned in "INDICES".'''
        
        # Setting up "tree_size"
        if (type(tree_size) is not ctypes.c_long): tree_size = ctypes.c_long(tree_size)
        
        # Setting up "layer"
        if (type(layer) is not ctypes.c_long): layer = ctypes.c_long(layer)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "found"
        if (type(found) is not ctypes.c_long): found = ctypes.c_long(found)
        
        # Setting up "starting_index"
        starting_index_present = ctypes.c_bool(True)
        if (starting_index is None):
            starting_index_present = ctypes.c_bool(False)
            starting_index = ctypes.c_long()
        else:
            starting_index = ctypes.c_long(starting_index)
        if (type(starting_index) is not ctypes.c_long): starting_index = ctypes.c_long(starting_index)
    
        # Call C-accessible Fortran wrapper.
        clib.c_level(ctypes.byref(tree_size), ctypes.byref(layer), ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(found), ctypes.byref(starting_index_present), ctypes.byref(starting_index))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, found.value

prune = prune()




# Get the version number from the setup file
import os

DIRECTORY = os.path.dirname(os.path.abspath(__file__))
ABOUT_DIR = os.path.join(DIRECTORY, "about")
VERSION_FILE = os.path.join(ABOUT_DIR, "version.txt")
if (os.path.exists(VERSION_FILE)):
    with open(VERSION_FILE) as f:
        __version__ = f.read().strip()
else:
    __version__ = "unknown"



# Define a set of test functions for approximation.
import numpy as np
# Imports for testing.
import os, time, json, gzip
from itertools import product
# Generating well-spaced points over the unit sphere (or ball) in any dimension.
from tlux.random import well_spaced_sphere, well_spaced_ball


REAL64_PRECISION = 2 ** (-52)
SQRT_REAL64_PRECISION = 2 ** (-26)


# Given a list of values, resize the list of values to the desired
#  size by linearly interpolating the percentiles of the data.
def linsize(values, new_size):
    return np.percentile(values, np.linspace(0, 100, new_size))


# Given a set of y values representing a 1D function, smooth the
#  function values by convolving a normal distirbution over the data.
def smoothy(x, y, stdev=1/30):
    if (stdev <= 0): return y
    x = np.asarray(x)
    y = np.asarray(y)
    new_y = np.zeros(y.shape)
    stdev *= max(SQRT_REAL64_PRECISION, x.max() - x.min())
    for i in range(y.size):
        # Generate convex weights from a normal distribution.
        weights = np.exp(
            -((x - x[i]) / stdev)**2 / 2
        ) / (stdev * np.sqrt(2*np.pi))
        # Adjust the weights at the ends of the interval to account 
        #   for clipping using the normal cumulative distirbution.
        left_tail_weight = (1.0 + np.math.erf((x[0]-x[i]) / (stdev * np.sqrt(2.0)))) / 2.0
        right_tail_weight = 1.0 - (1.0 + np.math.erf((x[-1]-x[i]) / (stdev * np.sqrt(2.0)))) / 2.0
        # Rescale the tail weights so their contribution is 
        #   proportional to their probability relative to the body.
        body_weight = 1.0 - left_tail_weight - right_tail_weight
        weight_scale = weights.sum() / max(SQRT_REAL64_PRECISION, body_weight)
        left_tail_weight *= weight_scale
        right_tail_weight *= weight_scale
        weights[0] += left_tail_weight
        weights[-1] += right_tail_weight
        # Make the weights convex.
        weights /= max(SQRT_REAL64_PRECISION, weights.sum())
        # Create the new y value (by convolving the normal distribution).
        new_y[i] = np.dot(y, weights)
    return new_y


# Produce a CDF fit of some values.
def cdf_fit(values, smooth=False, smooth_n=200, stdev=1/10, eps=SQRT_REAL64_PRECISION):
    assert len(values) > 0, "Provided values array has length 0."
    cdf_x = np.array(values)
    cdf_x.sort()
    # Ensure that all x are at least "eps" apart from each other.
    for i in range(len(cdf_x)-1):
        cdf_x[i+1] = max(cdf_x[i+1], cdf_x[i]+eps)
    cdf_y = np.linspace(0, 1, cdf_x.size)
    # Smooth the fit, if desired.
    if (smooth and (smooth_n > 0)):
        new_x = linsize(cdf_x, smooth_n)
        new_y = np.interp(new_x, cdf_x, cdf_y)
        new_y = smoothy(new_x, new_y, stdev=stdev)
        new_y -= new_y.min()
        new_y /= new_y.max()
        cdf_x, cdf_y = new_x, new_y
    # Generate the fit and its inverse.
    min_max = [cdf_x.min(), cdf_x.max()]
    fit = lambda *x: np.interp(x[0], cdf_x, cdf_y) if (len(x) > 0) else min_max
    fit.inverse = lambda y: np.interp(y, cdf_y, cdf_x)
    # Generate the approximate deirvative of the fit.
    x_widths = (cdf_x[1:] - cdf_x[:-1])
    mid_slopes_x = (cdf_x[:-1] + cdf_x[1:]) / 2
    mid_slopes_y = (cdf_y[1:] - cdf_y[:-1]) / x_widths
    slopes_y = cdf_x.copy()
    slopes_y[1:-1] = (mid_slopes_y[1:] + mid_slopes_y[:-1]) / 2
    slopes_y[0] = mid_slopes_y[0]
    slopes_y[-1] = mid_slopes_y[-1]
    # Normalize the area under the curve to be 1.
    slope_area = x_widths * (slopes_y[1:] + slopes_y[:-1]) / 2
    slopes_y /= slope_area.sum()
    fit.derivative = lambda x: np.interp(x, cdf_x, slopes_y)
    # Return the fit.
    return fit


# Test the CDF fit function.
def _test_linsize_smoothy_cdf_fit():
    from util.plot import Plot
    np.random.seed(0)
    # Generate a piecewise linear fit of a uniform distribution.
    data = np.random.uniform(size=(10000))
    cdf = cdf_fit(data)
    p = Plot()
    p.add_histogram("data", data, color=(0,0,0,0.3))
    p.add_func("CDF fit", cdf, cdf(), color=1)
    p.add_func("CDF inverse", cdf.inverse, [0, 1], color=2)
    p.show(show=False)
    # Generate a smoothed fit of a normal distribution.
    data = np.random.normal(size=(100))
    cdf = cdf_fit(data, smooth=True)
    p = Plot()
    p.add_histogram("data", data, color=(0,0,0,0.3))
    p.add_func("CDF fit", cdf, cdf(), color=1)
    p.add_func("CDF inverse", cdf.inverse, [0, 1], color=2)
    p.add_func("PDF", cdf.derivative, cdf(), color=0, line_width=0, fill="tozeroy")
    p.show(append=True)


# Return a random CDF.
def rand_cdf(nodes=3, power=1.0, smooth=200):
    # Generate random steps in the x and y direction (that sum to 1).
    cdf_x = np.linspace(0, 1, nodes+2)
    # Randomly assign a probability to each bin (this does not produce diverse CDFs).
    cdf_y = np.random.random(size=cdf_x.size) ** power
    cdf_y[0] = 0.0
    cdf_y /= cdf_y.sum()
    cdf_y = cdf_y.cumsum()
    # Smooth the CDF y values if desired.
    if smooth > 0:
        smooth += (smooth + 1) % 2 # Make sure "smooth" is odd.
        new_x = linsize(cdf_x, smooth)
        new_y = np.interp(new_x, cdf_x, cdf_y)
        new_y = smoothy(new_x, new_y)
        new_y -= new_y.min()
        new_y /= new_y.max()
        # Assign the new CDF points.
        cdf_x, cdf_y = new_x, new_y
    # Generate the CDF fit and return it.
    cdf = lambda x: np.interp(x, cdf_x, cdf_y)
    cdf.inverse = lambda x: np.interp(x, cdf_y, cdf_x)
    cdf.x = cdf_x
    cdf.y = cdf_y
    return cdf


# Verify that the CDF functions generated have nice behaviors.
def _test_rand_cdf():
    from util.stats import plot_percentiles
    from util.plot import Plot
    p = Plot()
    # Generate samples from a lot of random CDFs.
    x_points = np.linspace(0, 1, 1000)
    y_points = []
    samples = 500
    for i in range(samples):
        print(f"\r{i:5d} : {samples:5d}", end="")
        y_points.append(
            rand_cdf()(x_points)
        )
    print(" "*50)
    percentiles = list(range(0,101,10))
    y_points = np.asarray(y_points).T
    # Plot all of the percentiles of those CDFs to see the
    #  distribution of distributions that is generated.
    p.color_num += 1
    p = plot_percentiles(p, "Random CDFs", x_points, y_points,
                         percentiles=percentiles)
    # Add some examples for good measure.
    for i in range(5):
        f = rand_cdf()
        p.add_func(f"Sample {i+1}", f, [0, 1])
    # Generate the plot.
    p.show()


# Compute the cosine of the 2-norm of the x points (row vectors).
def cos_norm(x, radius=1, peaks=2, power=1):
    return np.cos(np.linalg.norm(x / radius, axis=1)**power
                  * max(0, peaks * 2 - 1) * np.pi).reshape((-1,1))


# Test the pure function that takes the cosine of the norm of x.
def _test_cos_norm():
    from util.plot import Plot
    p = Plot()
    n = 5000
    d = 2
    x = well_spaced_ball(n, d)
    y = cos_norm(x)
    p.add("cos 2-norm", *x.T, y[:,0], use_gradient=True)
    # p.add_func("numpy", x1_cdf, x1_cdf())
    # p.add_func("lcg", x2_cdf, x2_cdf())
    p.show(z_range=[-3,5])


# Apply a linear skew to a space such that the largest singular value
#  of the transformation is 1 and the smallest singular value is 'condition'.
def skew(f, d, condition=0.3):
    # Construct a random skew matrix.
    skew_matrix = well_spaced_sphere(d, d).T
    # Enforce the desired 2-norm properties (only shrink by at most 'condition').
    u, s, vt = np.linalg.svd(skew_matrix)
    s /= s[0]
    s *= np.linspace(1, condition / s[-1], d)
    skew_matrix = np.matmul(u * s, vt.T)
    # Construct the skew-wrapped function and return.
    def skewed_f(x, t=skew_matrix):
        return f(np.matmul(x, t))
    return skewed_f

# Test the skew and visually inspect the results are intuitively correct.
def _test_skew():
    f = cos_norm
    f_skew = skew(f, d=2)
    x = well_spaced_ball(10000, 2)
    from util.plot import Plot
    p = Plot()
    p.add("f", *x.T, f(x)[:,0], color=1, shade=True, marker_line_width=1)
    p.add("f_skew", *x.T, f_skew(x)[:,0], use_gradient=True)
    p.show()


# Apply a random CDF to the input components of `f` such that the min
#  and max values along each axis do not change. Make sure the axis 
#  aligned distribution functions are such that they never increase
#  the 2-norm of the points.
def redistribute(f, d, min_val=-1.0, max_val=1.0):
    # Generate CDF's that satisfy the necessary properties.
    n = 50
    cdfs = []
    for i in range(d):
        # ------------------------------------------------------------
        lx = np.linspace(min_val, 0, n)
        ly = np.asarray([
            np.linspace(0, np.pi, n) + np.sin(np.linspace(0, np.pi, n)),
            np.linspace(0, 2*np.pi, n) + np.sin(np.linspace(0, 2*np.pi, n)),
            np.linspace(0, 3*np.pi, n) + np.sin(np.linspace(0, 3*np.pi, n)),
            np.linspace(0, 4*np.pi, n) + np.sin(np.linspace(0, 4*np.pi, n)),
            np.linspace(0, 5*np.pi, n) + np.sin(np.linspace(0, 5*np.pi, n)),
            np.linspace(0, 6*np.pi, n) + np.sin(np.linspace(0, 6*np.pi, n)),
        ]).T
        ly /= ly.max(axis=0)
        # Randomly weight the basis functions.
        w = np.random.random(size=(ly.shape[1],))
        w /= w.sum()
        # Update the weights to favor values above the diagonal.
        vals = np.dot(ly,w)
        diff_vals = vals - np.linspace(0,1,n)
        y0_vals = ly[:,0]
        gap = y0_vals - vals
        ratio = np.where(gap != 0,
                         -diff_vals / np.where(gap != 0, gap, 1.0),
                         0.0)
        w[:] *= (1.0 - ratio.max())
        w[0] += ratio.max()
        # Finalize the left y values.
        ly = np.dot(ly, w) - 1.0
        # ------------------------------------------------------------
        # Make the right y values with the same method (flipped).
        rx = np.linspace(0, max_val, n)
        ry = np.asarray([
            np.linspace(0, np.pi, n) + np.sin(np.linspace(0, np.pi, n)),
            np.linspace(0, 2*np.pi, n) + np.sin(np.linspace(0, 2*np.pi, n)),
            np.linspace(0, 3*np.pi, n) + np.sin(np.linspace(0, 3*np.pi, n)),
            np.linspace(0, 4*np.pi, n) + np.sin(np.linspace(0, 4*np.pi, n)),
            np.linspace(0, 5*np.pi, n) + np.sin(np.linspace(0, 5*np.pi, n)),
            np.linspace(0, 6*np.pi, n) + np.sin(np.linspace(0, 6*np.pi, n)),
        ]).T
        ry /= ry.max(axis=0)
        # Randomly weight the basis functions.
        w = np.random.random(size=(ry.shape[1],))
        w /= w.sum()
        # Update the weights to favor values above the diagonal.
        vals = np.dot(ry,w)
        diff_vals = vals - np.linspace(0,1,n)
        y0_vals = ry[:,0]
        gap = y0_vals - vals
        ratio = np.where(gap != 0,
                         -diff_vals / np.where(gap != 0, gap, 1.0),
                         0.0)
        w[:] *= (1.0 - ratio.max())
        w[0] += ratio.max()
        # Finalize the right y values.
        ry = np.dot(ry, w)
        # Flip the values and reorder them.
        ry = (1.0 - ry)[::-1]
        # ------------------------------------------------------------
        # Create and store the fit function.
        x = np.concatenate((lx, rx), axis=0)
        y = np.concatenate((ly, ry), axis=0)
        fit = lambda z, x=x, y=y: np.interp(z, x, y)
        fit.inverse = lambda z, x=x, y=y: np.interp(z, y, x)
        fit.x = x
        fit.y = y
        cdfs.append(fit)
    # Generate the axis-aligned redistribution and return.
    def redistributed_f(x, cdfs=cdfs):
        new_x = x.copy()
        for i in range(d):
            new_x[:,i] = cdfs[i](x[:,i])
        return f(new_x)
    return redistributed_f

def _test_redistribute():
    f = cos_norm
    f_rd = redistribute(f, 2)
    x = well_spaced_ball(10000, 2)
    from util.plot import Plot
    p = Plot()
    p.add("f", *x.T, f(x)[:,0], color=1, shade=True, marker_line_width=1)
    p.add("f_rd", *x.T, f_rd(x)[:,0], use_gradient=True)
    p.show()


# A pure approximation problem. Should serve as a viable test for
# approximation algorithms. Returns a test approximation function that
# returns two arrays, x and y, given x inputs for approximation.
# 
#   d (integer)
#      Dimension of the function input.
#   df (optional integer)
#      Dimension of the function output. Defaults to 1.
#   df_intrinsic (optional integer)
#      Intrinsic dimension of the output. I.e., number of principal
#      components with nonzero variance. Default is max(1, df // 2).
#   shift (real)
#      Magnitude of the random shift applied to the input data for
#      each intrinsic dimension of output.
#   distort (optional boolean)
#      If True, apply a random distribution transformation that 
#      distorts the axis-aligned input point distribution. 
#      Defaults to False.
#   skew (optional boolean)
#      If True, apply a random linear transformation that rotates and
#      skews the input points. Defaults to False.
#   ortho (optional boolean)
#      This flag determines whether orthogonality is enforced over the
#      outputs of the underlying function in the intrinsic dimension,
#      necessary for orthogonality when distort=True of skew=True.
#      Default value is True.
#      WARNING: Setting this to True will result in different spatial
#               function evaluations depending on the number of inputs.
# 
def pure(d, df=1, df_intrinsic=None, shift=0.2, complexity=2,
         distort=False, skew=False, ortho=True, seed=0):
    np.random.seed(seed)
    # Set the intrinsic dimension of the output if it was not provided.
    if (df_intrinsic is None):
        df_intrinsic = max(1, df // 2)
    # If output is in higher dimension than the intrinsic, it will
    #  be lifted with a linear projection onto well spaced vectors.
    projection = None
    if (df != df_intrinsic):
        projection = well_spaced_sphere(df, df_intrinsic).T
    # Add some centered shift to the input data of the function.
    shift_vecs = np.vstack((
        np.zeros((1, d)),
        well_spaced_sphere(df_intrinsic-1, d) * shift
    ))
    # Apply a random distribution distortion to all input components.
    cdfs = None
    if (distort):
        cdfs = [rand_cdf() for i in range(d)]
    # Apply a random linear projection to x to "skew" it.
    if (skew):
        skew = well_spaced_sphere(d, d).T
    else:
        skew = None
    # Create a function with orthogonal outputs (in the intrinsic dimension).
    def pure_func(x, complexity=complexity, shift=shift, shift_vecs=shift_vecs,
                  projection=projection, cdfs=cdfs, skew=skew, ortho=ortho):
        # Check quality of x input.
        assert (len(x.shape) == 2), f"Expected x with shape (n,{d}) but received x with shape {x.shape}."
        assert (x.shape[1] == d), f"Expected row vectors in x with dimension {d} but received x with shape {x.shape}."
        # Apply a distribution distortion to input components
        #   assuming ideal x component range is [-1,1].
        if (cdfs is not None):
            x = x.copy()
            for i in range(d):
                x[:,i] = cdfs[i].inverse((x[:,i] + 1) / 2) * 2 - 1
        # Apply a linear skew to the input.
        if (skew is not None):
            x = np.matmul(x, skew)
        # Compute a function over inputs that has orthogonal outputs.
        y = [cos_norm(x + shift_vecs[i], radius=1+shift, peaks=complexity) 
             for i in range(df_intrinsic)]
        y = np.asarray(y).reshape((df_intrinsic, -1)).T
        # Make the y outputs orthogonal in the intrinsic dimension.
        if (ortho and (df_intrinsic > 1)):
            old_y, (y, _) = y, np.linalg.qr(y)
            numerator = 1.0 if (np.dot(old_y[:,0], y[:,0]) >= 0) else -1.0
            y_scale = numerator / np.abs(y[:,0]).max()
            y *= y_scale
        # Project the output into a higher dimension.
        if (projection is not None):
            y = np.matmul(y, projection)
        # Return the final function output.
        return x, y
    # Reset the random seed for later processes.
    np.random.seed()
    # Return the pure function.
    return pure_func


# Test the random uniform by plotting the CDF.
def _test_pure(normalize=False):
    from util.plot import Plot
    p = Plot()
    n = 500
    d = 2
    df = 1
    f = pure(d, df=df)
    x = well_spaced_ball(n, d)
    new_x, y = f(x)
    # Zero mean and unit variance the data like the neural network would.
    if normalize:
        x -= x.mean(axis=0)
        x /= x.var(axis=0)
        y -= y.mean(axis=0)
        y /= y.var(axis=0)
    # Visualize the pure function components.
    for i in range(df):
        p.add(f"pure {i} x", *x.T, y[:,i] * 0.0 + y[:,i].min() - 0.7, marker_size=2, color=(200,200,200),
              marker_line_color=(0,0,0), marker_line_width=1, group=i, show_in_legend=False)
        p.add(f"pure {i}", *x.T, y[:,i], marker_size=4, use_gradient=True, marker_line_width=1, group=i)
    p.show(z_range=[-3,5])


# Create a function that samples an underlying surface. The ground
#  truth prediction task has two parts, identify the type of
#  underlying surface and identify the center of that type.
def sampled(d, df=4, dp=2, df_intrinsic=None,
            max_shift=0.5, power=3, return_info=False, seed=0):
    # By default, make all of the "modes" that can be observed maximally different.
    if (df_intrinsic is None):
        df_intrinsic = 2 * df
    # Define the output dimension of the pure function based on the total output dimension.
    dp = min(dp, df)
    fdf = max(1, df - dp)
    # Generate a "truth function" that creates orthogonal output
    #  components that can be combined to form "category predictions".
    f = pure(dp, df=fdf, df_intrinsic=df_intrinsic, seed=seed)
    # Generate positions for all of the dimensions in some space.
    np.random.seed(seed)
    positions = well_spaced_ball(d, dp)
    np.random.seed()
    # Rescale the function output so that the final dimension can be matched.
    def sampled_func(inputs, d=d, dp=dp, df=df, fdf=fdf, max_shift=max_shift,
                     power=power, f=f, positions=positions, seed=seed):
        assert len(inputs.shape) == 2, f"Expected input array to have shape (n,{d}), received {inputs.shape}."
        assert inputs.shape[1] == d,   f"Expected input array to have shape (n,{d}), received {inputs.shape}."
        np.random.seed(seed)
        # Get the number of points.
        n = inputs.shape[0]
        # Assign random center offsets to all of the points (the y values).
        offsets = well_spaced_ball(n, dp) * max_shift
        # Generate a set of evaluation points (at all positions, for all offsets).
        x = np.concatenate([(positions - o) for o in offsets])
        # Evaluate the function at all positions, this will produce
        #   fdf channels of output for every (position,offset) pair.
        new_x, y = f(x)
        # Generate a set of weights, this will be used to blend the modes
        #   of the output from the function into single "observations".
        #   Ideally, the "most observable" output is the predicted one.
        if (fdf + dp == df):
            weights = (well_spaced_sphere(n, fdf) + 1) ** power
        else:
            weights = np.ones((n, fdf))
        weights = (weights.T / np.sum(weights, axis=1)).T
        # Overwrite the inputs, collapsing the observations using the weights.
        for i in range(n):
            inputs[i,:] = np.dot(y[i*d:(i+1)*d], weights[i])
        # The outputs are the offset terms and weights concatenated,
        #  both are something that should be predicted by an 
        #  approximator. Unless there is not room in the output.
        if (fdf + dp == df):
            outputs = np.concatenate((offsets, weights), axis=1)
        else:
            outputs = offsets
        # Return the pairs of points.
        np.random.seed()
        return inputs, outputs
    # Store some configurations in accessible attributes.
    sampled_func.f = f
    sampled_func.positions = positions
    # Return the function.
    return sampled_func


# Test the function that mimics the sampling behavior.
def _test_sampled():
    from util.plot import Plot
    n = 10
    d = 3000
    f = sampled(d, return_info=True)
    positions = f.positions
    x = well_spaced_sphere(n, d)
    _, y = f(x)
    p = Plot()
    for i in range(n):
        p.add(f"post {i+1}", [y[i,0],y[i,0]], [y[i,1], y[i,1]], [-2, 2],
              mode="lines", line_width=4, color=1, group=i, show_in_legend=False)
        p.add(f"Offset {i+1} ({y[i,0]:.3f}, {y[i,1]:.3f}) [{y[i,2:].round(3)}]",
              *positions.T, x[i], use_gradient=True,
              marker_size=4, marker_line_width=1, group=i)
    p.show(z_range=[-3,5])


# Test all functions.
def _test_all():
    _test_linsize_smoothy_cdf_fit()
    _test_rand_cdf()
    _test_cos_norm()
    _test_skew()
    _test_redistribute()
    _test_pure()
    _test_sampled()


# Define functions that map from a configuration to identifier strings.
def model_str(M):
    return str(M).split("'")[1].replace(".", "-").split("-")[-1]
def function_str(f):
    return str(str(f).split()[1:2])[2:-2]
def config_str(config):
    M, f, n, d, df, ds, ns, s, test_size, func_seed, model_seed = config
    return f"{func_seed:02d}-{model_seed:02d} {function_str(f):8s} {n:5d} {d:5d} {df} {model_str(M)}-{ds:02d}-{ns:02d} ({s} {test_size})"
def test_id_str(config):
    M, f, n, d, df, ds, ns, s, test_size, func_seed, model_seed = config
    return f"{model_seed:04d}-{func_seed:04d}-{function_str(f)}"

# Given a model class, instantiate and test various problems.
def test_model_class(M, suffix="", test_size=1000,
                     func_seeds=[1,2], #list(range(1,5+1)),
                     model_seeds=list(range(1,10+1)),
                     funcs=[pure, sampled],
                     num_points=[50, 100, 500],
                     dimension_in=[2, 10, 100],
                     dimension_out=[1, 5],
                     model_internal_dimension=[32],
                     model_internal_states=[8],
                     model_fit_steps=[3000]):
    # Run a single test given a model class M. Return the results.
    def run_test(config):
        # Extract the configuration variables.
        M, f, n, d, df, ds, ns, s, test_size, func_seed, model_seed = config
        np.random.seed(func_seed)
        # Generate random test data.
        func = f(d, df=df, seed=func_seed)
        x = well_spaced_ball(n+test_size, d)
        new_x, y = func(x)
        # Componentwise zero mean and unit variance all data.
        x_mean = x.mean(axis=0)
        x_stdev = x.std(axis=0)
        y_mean = y.mean(axis=0)
        y_stdev = y.std(axis=0)
        x = (x - x_mean) / np.where(x_stdev > 0, x_stdev, 1.0)
        y = (y - y_mean) / np.where(y_stdev > 0, y_stdev, 1.0)
        # Assign training and testing points.
        train_inds = np.arange(n+test_size)
        np.random.shuffle(train_inds)
        train_inds, test_inds = train_inds[:n], train_inds[n:]
        # Initialize the model.
        model = M(di=d, do=df, ds=ds, ns=ns, seed=model_seed)
        # Compute the error of the model before the fit operation.
        output = model(x[test_inds])
        prefit_error = ((output - y[test_inds])**2).sum(axis=1).mean()
        # Fit the model.
        fit_start = time.time()
        model.fit(x[train_inds], y[train_inds], new_model=False,
                  normalize_x=False, normalize_y=False, steps=s,
                  validation_size=0, early_stop=True, keep_best=False,
                  num_threads=2)
        fit_time = time.time() - fit_start
        # Predict with the model.
        predict_start = time.time()
        output = model(x[test_inds])
        predict_time = time.time() - predict_start
        # Measure the error (mean squared and percentiles).
        error = ((output - y[test_inds])**2).sum(axis=1)
        error_deciles = np.percentile(error, list(range(0,101,10)))
        # Return the test results.
        return {
            "model": model_str(M),
            "function": function_str(f),
            "func_seed": func_seed,
            "model_seed": model_seed,
            "n": n,
            "d": d,
            "df": df,
            "ds": ds,
            "ns": ns,
            "steps": s,
            "test_size": test_size,
            "fit_time": fit_time,
            "predict_time": predict_time,
            "initial_mean_squared_error": float(prefit_error),
            "mean_squared_error": float(error.mean()),
            "error_deciles": error_deciles.tolist(),
            "fit_record": getattr(model, "record", np.zeros(0)).tolist(),
        }
        
    # Load in any existing data that's already been collected.
    data_dir = os.path.join("data", f"testing_{model_str(M)}{suffix}")
    os.makedirs(data_dir, exist_ok=True)
    test_configs = set()
    for test_id in os.listdir(data_dir):
        with gzip.open(os.path.join(data_dir, test_id), "rb") as f:
            test_results = json.loads(str(f.read(), "utf8"))
        test_configs |= set(test_results)
    # Make all of the configurations for testing, skipping already covered tests.
    configs = []
    for (fs, ms, f, n, d, df, ds, ns, s) in product(
            func_seeds, model_seeds, funcs, num_points, dimension_in, 
            dimension_out, model_internal_dimension, 
            model_internal_states, model_fit_steps, 
    ):
        config = (M, f, n, d, df, ds, ns, s, test_size, fs, ms)
        conf_str = config_str(config)
        if (conf_str in test_configs):
            print(conf_str)
        else:
            configs.append(config)
    # Execute all of the tests.
    previous_time = time.time()
    previous_id = None
    test_results = {}
    for config in configs:
        # If the ID of this test doesn't match the previous ID, then
        #  save all previous results to a file and reset the result list..
        test_id = test_id_str(config)
        if (test_id != previous_id):
            if (previous_id is not None):
                with gzip.open(os.path.join(data_dir, f"{previous_id}.json.gz"), "wb") as f:
                    f.write(bytes(json.dumps(test_results), "utf8"))
            # Check for existing data from the new test config, load it if it exists.
            test_data_path = os.path.join(data_dir, f"{test_id}.json.gz")
            if os.path.exists(test_data_path):
                with gzip.open(test_data_path, "rb") as f:
                    test_results = json.loads(str(f.read(), "utf8"))
            else:
                test_results = {}
            # Reset the previous ID and the previous time.
            previous_id = test_id
            previous_time = time.time()
        # Run this new test.
        result = run_test(config)
        # Store test result.
        conf_str = config_str(config)
        test_results[conf_str] = result
        # Print an update.
        print(f"{conf_str}  {time.ctime()}  ({time.time() - previous_time:0.2f} seconds)")
        previous_time = time.time()
    # Save the last test iteration.
    with gzip.open(os.path.join(data_dir, f"{previous_id}.json.gz"), "wb") as f:
        f.write(bytes(json.dumps(test_results), "utf8"))



# Visualize the test results for a model class.
def view_model_class(M, suffix=""):
    from util.data import Data
    # Declare a file for the data to be read quickly.
    data_file = f"data_{model_str(M)}{suffix}.pkl"
    if (not os.path.exists(data_file)):
        print("Reading data from directory..")
        # Load in any existing data that's already been collected.
        data_dir = os.path.join("data", f"testing_{model_str(M)}{suffix}")
        names = ['n', 'd', 'df', 'function', 'model', 'ds', 'ns',
                 'steps', 'test_size', 'func_seed', 'model_seed',
                 'fit_time', 'predict_time',
                 'initial_mean_squared_error', 'mean_squared_error',
                 'error_deciles', 'fit_record']
        configs = set(names[:9])
        trials = set(names[9:11])
        # Read the data.
        d = Data(names=names)
        for test_id in sorted(os.listdir(data_dir)):
            with gzip.open(os.path.join(data_dir, test_id), "rb") as f:
                test_results = json.loads(str(f.read(), "utf8"))
            for conf_str in sorted(test_results):
                d.append([test_results[conf_str][n] for n in names])
        # Stack the data across the axes that are repeated for each trial.
        to_stack = [n for n in d.names if n not in configs]
        print("  reorganizing data, stacking across trials..")
        d.stack(to_stack)
        print("  deleting columns that only have 1 unique value..")
        import pickle
        hash_func = lambda v: hash(pickle.dumps(v))
        num_unique = {n:len(set(map(hash_func, d[n]))) for n in d.names}
        for n in list(d.names):
            if (num_unique[n] == 1):
                print("  ", n, d[n][0])
                d.pop(n)
        print("  saving data for faster loading..")
        d.save(data_file)
    else:
        print("Loading data from file..")
        d = Data.load(data_file)
    
    print()
    print(d)
    print()

    # Get the names of the columns that are for test configuration.
    conf_cols = [n for n,t in zip(d.names, d.types) if t is not list][::-1]
    print("conf_cols: ", conf_cols)
    trial_cols = [n for n,t in zip(d.names, d.types) if t is list]
    print("trial_cols: ", trial_cols)
    print()
    d.sort(key=lambda row: tuple(row[conf_cols]))

    # Create plots for each config.
    from util.plot import Plot
    from util.stats.plotting import plot_percentiles
    for i,test in enumerate(d):
        config = test[conf_cols]
        conf_str = " ".join(list(map(str, config)))
        print("", conf_str, end="   ")
        # Create a visual of the model training error and test error.
        p = Plot(conf_str, "step", "mean squared error")
        # Plot the percentiles of the training error.
        records = np.asarray(test["fit_record"]).T
        step = np.arange(1, 1+records.shape[1])
        p = plot_percentiles(p, "MSE", step, records[0,:,:], color=1)
        #   plot the validation error if that was recorded
        if (np.linalg.norm(records[0,:,:]-records[1,:,:]) > 0.01):
            p = plot_percentiles(p, "Validation MSE", step, records[1,:,:], color=0)
        # Plot the testing error distribution.
        def sideways_pdf(name, values, color=0):
            cdf = cdf_fit(values, smooth=True)
            pdf_x = cdf.inverse(np.linspace(0, 1, 1000))
            pdf_y = cdf.derivative(pdf_x)
            #    reorient the CDF to be on the far right
            pdf_x, pdf_y = pdf_y, pdf_x
            pdf_x -= min(pdf_x)
            pdf_x /= -max(pdf_x)
            pdf_x -= min(pdf_x)
            pdf_x *= (max(step) - min(step)) / 20
            pdf_x += max(step)
            pdf_x = pdf_x.tolist()
            pdf_y = pdf_y.tolist()
            pdf_x = [max(pdf_x)] + pdf_x + [max(pdf_x)]
            pdf_y = [pdf_y[0]] + pdf_y + [pdf_y[-1]]
            # Plot the distributions.
            p.add(f"{name}", pdf_x, pdf_y, mode="lines", color=color,
                  fill="toself", line_width=0, group="test cdf")
            pdf_x = pdf_x[1:-1]
            pdf_y = pdf_y[1:-1]
            p.add(f"{name} Line", pdf_x, pdf_y, mode="lines", color=(0,0,0,0.8),
                  line_width=0.5, show_in_legend=False, group="test cdf")
        sideways_pdf("Prefit Error", test["initial_mean_squared_error"], color=0)
        sideways_pdf("Test Error", test["mean_squared_error"], color=3)
        # Show the plot (only the first and the last).
        p.show(append=True, show=(i==0 or i==len(d)-1))


from util.approximate import PLRM

# suffix = "_validation-based"
# suffix = "_no-validation-early-stopping"
# suffix = "_no-validation-no-early-stopping"
suffix = "_no-validation-early-stopping-no-keep-best"

# Run tests for a model class.
test_model_class(PLRM, suffix=suffix)

from notifications import send_email
send_email("", "Done testing")

# View the test results for a model class.
view_model_class(PLRM, suffix=suffix)


# TODO:
# 
#  - restructure test functions, make the cos(norm()) function pure
#    without quadratic modifications (distribution shift of input can
#    have the same effect of skewing space).
# 
#  - given two suffixes, produce the "probability best" plots
#    for each test configuration
# 
#  - make sure that seeded initializations and functions are correctly
#    making everything deterministic (looks like they are not)
# 
#  - make skew and distort into decorators, remove from body of 'pure'
# 


# --------------------------------------------------------------------

# "positional" function is the pure Cosine of 2-norm
# "apositional" function is the spatially sampled value of pure function
# "category" defines which function is being sampled
# 



'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.

! A nearest neighbor tree structure that picks points on the convex hull and
!  splits regions in half by the 2-norm distance to the median child.
!  Construction is parallelized for shared memory architectures with OpenMP,
!  and querying is parallelized over batched query points (but serial for
!  a single query). In addition a nearest neighbor query budget can be
!  provided to generate approximate results, with the guarantee that exact
!  nearest neighbors will be found given a budget greater than the logarithm
!  base two of the number of points in the tree.
'''

import os
import ctypes
import platform
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "ball_tree." + platform.machine() + ".so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-fPIC', '-shared', '-O3', '-fopenmp']
_ordered_dependencies = ['libomp.dylib', 'swap.f90', 'prune.f90', 'fast_select.f90', 'fast_sort.f90', 'ball_tree.f90', 'ball_tree_c_wrapper.f90']
_symbol_files = ['libomp.dylib']# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the prerequisite symbols for the compiled code.
for _ in _symbol_files:
    _ = ctypes.CDLL(os.path.join(_this_directory, _), mode=ctypes.RTLD_GLOBAL)
# Try to import the existing object. If that fails, recompile and then try.
try:
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


class ball_tree:
    ''''''

    # Declare 'max_copy_bytes'
    def get_max_copy_bytes(self):
        max_copy_bytes = ctypes.c_long()
        clib.ball_tree_get_max_copy_bytes(ctypes.byref(max_copy_bytes))
        return max_copy_bytes.value
    def set_max_copy_bytes(self, max_copy_bytes):
        max_copy_bytes = ctypes.c_long(max_copy_bytes)
        clib.ball_tree_set_max_copy_bytes(ctypes.byref(max_copy_bytes))
    max_copy_bytes = property(get_max_copy_bytes, set_max_copy_bytes)

    # Declare 'number_of_threads'
    def get_number_of_threads(self):
        number_of_threads = ctypes.c_int()
        clib.ball_tree_get_number_of_threads(ctypes.byref(number_of_threads))
        return number_of_threads.value
    def set_number_of_threads(self, number_of_threads):
        number_of_threads = ctypes.c_int(number_of_threads)
        clib.ball_tree_set_number_of_threads(ctypes.byref(number_of_threads))
    number_of_threads = property(get_number_of_threads, set_number_of_threads)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine CONFIGURE
    
    def configure(self, num_threads=None, max_levels=None, nested=None):
        '''! Configure OpenMP parallelism for this ball tree code.'''
        
        # Setting up "num_threads"
        num_threads_present = ctypes.c_bool(True)
        if (num_threads is None):
            num_threads_present = ctypes.c_bool(False)
            num_threads = ctypes.c_int()
        else:
            num_threads = ctypes.c_int(num_threads)
        if (type(num_threads) is not ctypes.c_int): num_threads = ctypes.c_int(num_threads)
        
        # Setting up "max_levels"
        max_levels_present = ctypes.c_bool(True)
        if (max_levels is None):
            max_levels_present = ctypes.c_bool(False)
            max_levels = ctypes.c_int()
        else:
            max_levels = ctypes.c_int(max_levels)
        if (type(max_levels) is not ctypes.c_int): max_levels = ctypes.c_int(max_levels)
        
        # Setting up "nested"
        nested_present = ctypes.c_bool(True)
        if (nested is None):
            nested_present = ctypes.c_bool(False)
            nested = ctypes.c_int()
        else:
            nested = ctypes.c_int(nested)
        if (type(nested) is not ctypes.c_int): nested = ctypes.c_int(nested)
    
        # Call C-accessible Fortran wrapper.
        clib.c_configure(ctypes.byref(num_threads_present), ctypes.byref(num_threads), ctypes.byref(max_levels_present), ctypes.byref(max_levels), ctypes.byref(nested_present), ctypes.byref(nested))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return (num_threads.value if num_threads_present else None), (max_levels.value if max_levels_present else None), (nested.value if nested_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine COMPUTE_SQUARE_SUMS
    
    def compute_square_sums(self, points, sq_sums):
        '''! Compute the square sums of a bunch of points (with parallelism).'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
    
        # Call C-accessible Fortran wrapper.
        clib.c_compute_square_sums(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return sq_sums

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine BUILD_TREE
    
    def build_tree(self, points, sq_sums, radii, medians, sq_dists, order, root=None, leaf_size=None):
        '''! Re-arrange elements of POINTS into a binary ball tree about medians.'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "medians"
        if ((not issubclass(type(medians), numpy.ndarray)) or
            (not numpy.asarray(medians).flags.f_contiguous) or
            (not (medians.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'medians' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            medians = numpy.asarray(medians, dtype=ctypes.c_float, order='F')
        medians_dim_1 = ctypes.c_long(medians.shape[0])
        
        # Setting up "sq_dists"
        if ((not issubclass(type(sq_dists), numpy.ndarray)) or
            (not numpy.asarray(sq_dists).flags.f_contiguous) or
            (not (sq_dists.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_dists' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_dists = numpy.asarray(sq_dists, dtype=ctypes.c_float, order='F')
        sq_dists_dim_1 = ctypes.c_long(sq_dists.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "root"
        root_present = ctypes.c_bool(True)
        if (root is None):
            root_present = ctypes.c_bool(False)
            root = ctypes.c_long()
        else:
            root = ctypes.c_long(root)
        if (type(root) is not ctypes.c_long): root = ctypes.c_long(root)
        
        # Setting up "leaf_size"
        leaf_size_present = ctypes.c_bool(True)
        if (leaf_size is None):
            leaf_size_present = ctypes.c_bool(False)
            leaf_size = ctypes.c_long()
        else:
            leaf_size = ctypes.c_long(leaf_size)
        if (type(leaf_size) is not ctypes.c_long): leaf_size = ctypes.c_long(leaf_size)
    
        # Call C-accessible Fortran wrapper.
        clib.c_build_tree(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(medians_dim_1), ctypes.c_void_p(medians.ctypes.data), ctypes.byref(sq_dists_dim_1), ctypes.c_void_p(sq_dists.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(root_present), ctypes.byref(root), ctypes.byref(leaf_size_present), ctypes.byref(leaf_size))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return points, sq_sums, radii, medians, sq_dists, order

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine NEAREST
    
    def nearest(self, points, k, tree, sq_sums, radii, medians, order, leaf_size, indices, dists, iwork, rwork, to_search=None, randomness=None):
        '''! Compute the K nearest elements of TREE to each point in POINTS.'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "k"
        if (type(k) is not ctypes.c_long): k = ctypes.c_long(k)
        
        # Setting up "tree"
        if ((not issubclass(type(tree), numpy.ndarray)) or
            (not numpy.asarray(tree).flags.f_contiguous) or
            (not (tree.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'tree' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            tree = numpy.asarray(tree, dtype=ctypes.c_float, order='F')
        tree_dim_1 = ctypes.c_long(tree.shape[0])
        tree_dim_2 = ctypes.c_long(tree.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "medians"
        if ((not issubclass(type(medians), numpy.ndarray)) or
            (not numpy.asarray(medians).flags.f_contiguous) or
            (not (medians.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'medians' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            medians = numpy.asarray(medians, dtype=ctypes.c_float, order='F')
        medians_dim_1 = ctypes.c_long(medians.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "leaf_size"
        if (type(leaf_size) is not ctypes.c_long): leaf_size = ctypes.c_long(leaf_size)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        indices_dim_2 = ctypes.c_long(indices.shape[1])
        
        # Setting up "dists"
        if ((not issubclass(type(dists), numpy.ndarray)) or
            (not numpy.asarray(dists).flags.f_contiguous) or
            (not (dists.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'dists' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            dists = numpy.asarray(dists, dtype=ctypes.c_float, order='F')
        dists_dim_1 = ctypes.c_long(dists.shape[0])
        dists_dim_2 = ctypes.c_long(dists.shape[1])
        
        # Setting up "iwork"
        if ((not issubclass(type(iwork), numpy.ndarray)) or
            (not numpy.asarray(iwork).flags.f_contiguous) or
            (not (iwork.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'iwork' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            iwork = numpy.asarray(iwork, dtype=ctypes.c_long, order='F')
        iwork_dim_1 = ctypes.c_long(iwork.shape[0])
        iwork_dim_2 = ctypes.c_long(iwork.shape[1])
        
        # Setting up "rwork"
        if ((not issubclass(type(rwork), numpy.ndarray)) or
            (not numpy.asarray(rwork).flags.f_contiguous) or
            (not (rwork.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'rwork' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            rwork = numpy.asarray(rwork, dtype=ctypes.c_float, order='F')
        rwork_dim_1 = ctypes.c_long(rwork.shape[0])
        rwork_dim_2 = ctypes.c_long(rwork.shape[1])
        
        # Setting up "to_search"
        to_search_present = ctypes.c_bool(True)
        if (to_search is None):
            to_search_present = ctypes.c_bool(False)
            to_search = ctypes.c_long()
        else:
            to_search = ctypes.c_long(to_search)
        if (type(to_search) is not ctypes.c_long): to_search = ctypes.c_long(to_search)
        
        # Setting up "randomness"
        randomness_present = ctypes.c_bool(True)
        if (randomness is None):
            randomness_present = ctypes.c_bool(False)
            randomness = ctypes.c_float()
        else:
            randomness = ctypes.c_float(randomness)
        if (type(randomness) is not ctypes.c_float): randomness = ctypes.c_float(randomness)
    
        # Call C-accessible Fortran wrapper.
        clib.c_nearest(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(k), ctypes.byref(tree_dim_1), ctypes.byref(tree_dim_2), ctypes.c_void_p(tree.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(medians_dim_1), ctypes.c_void_p(medians.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(leaf_size), ctypes.byref(indices_dim_1), ctypes.byref(indices_dim_2), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(dists_dim_1), ctypes.byref(dists_dim_2), ctypes.c_void_p(dists.ctypes.data), ctypes.byref(iwork_dim_1), ctypes.byref(iwork_dim_2), ctypes.c_void_p(iwork.ctypes.data), ctypes.byref(rwork_dim_1), ctypes.byref(rwork_dim_2), ctypes.c_void_p(rwork.ctypes.data), ctypes.byref(to_search_present), ctypes.byref(to_search), ctypes.byref(randomness_present), ctypes.byref(randomness))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, dists, iwork, rwork

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine PT_NEAREST
    
    def pt_nearest(self, point, k, tree, sq_sums, radii, medians, order, leaf_size, indices, dists, randomness, checks=None, found=None, pt_ss=None, d_root=None):
        '''! Compute the K nearest elements of TREE to each point in POINTS.'''
        
        # Setting up "point"
        if ((not issubclass(type(point), numpy.ndarray)) or
            (not numpy.asarray(point).flags.f_contiguous) or
            (not (point.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'point' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            point = numpy.asarray(point, dtype=ctypes.c_float, order='F')
        point_dim_1 = ctypes.c_long(point.shape[0])
        
        # Setting up "k"
        if (type(k) is not ctypes.c_long): k = ctypes.c_long(k)
        
        # Setting up "tree"
        if ((not issubclass(type(tree), numpy.ndarray)) or
            (not numpy.asarray(tree).flags.f_contiguous) or
            (not (tree.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'tree' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            tree = numpy.asarray(tree, dtype=ctypes.c_float, order='F')
        tree_dim_1 = ctypes.c_long(tree.shape[0])
        tree_dim_2 = ctypes.c_long(tree.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "medians"
        if ((not issubclass(type(medians), numpy.ndarray)) or
            (not numpy.asarray(medians).flags.f_contiguous) or
            (not (medians.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'medians' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            medians = numpy.asarray(medians, dtype=ctypes.c_float, order='F')
        medians_dim_1 = ctypes.c_long(medians.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "leaf_size"
        if (type(leaf_size) is not ctypes.c_long): leaf_size = ctypes.c_long(leaf_size)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "dists"
        if ((not issubclass(type(dists), numpy.ndarray)) or
            (not numpy.asarray(dists).flags.f_contiguous) or
            (not (dists.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'dists' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            dists = numpy.asarray(dists, dtype=ctypes.c_float, order='F')
        dists_dim_1 = ctypes.c_long(dists.shape[0])
        
        # Setting up "randomness"
        if (type(randomness) is not ctypes.c_float): randomness = ctypes.c_float(randomness)
        
        # Setting up "checks"
        checks_present = ctypes.c_bool(True)
        if (checks is None):
            checks_present = ctypes.c_bool(False)
            checks = ctypes.c_long()
        else:
            checks = ctypes.c_long(checks)
        if (type(checks) is not ctypes.c_long): checks = ctypes.c_long(checks)
        
        # Setting up "found"
        found_present = ctypes.c_bool(True)
        if (found is None):
            found_present = ctypes.c_bool(False)
            found = ctypes.c_long()
        else:
            found = ctypes.c_long(found)
        if (type(found) is not ctypes.c_long): found = ctypes.c_long(found)
        
        # Setting up "pt_ss"
        pt_ss_present = ctypes.c_bool(True)
        if (pt_ss is None):
            pt_ss_present = ctypes.c_bool(False)
            pt_ss = ctypes.c_float()
        else:
            pt_ss = ctypes.c_float(pt_ss)
        if (type(pt_ss) is not ctypes.c_float): pt_ss = ctypes.c_float(pt_ss)
        
        # Setting up "d_root"
        d_root_present = ctypes.c_bool(True)
        if (d_root is None):
            d_root_present = ctypes.c_bool(False)
            d_root = ctypes.c_float()
        else:
            d_root = ctypes.c_float(d_root)
        if (type(d_root) is not ctypes.c_float): d_root = ctypes.c_float(d_root)
    
        # Call C-accessible Fortran wrapper.
        clib.c_pt_nearest(ctypes.byref(point_dim_1), ctypes.c_void_p(point.ctypes.data), ctypes.byref(k), ctypes.byref(tree_dim_1), ctypes.byref(tree_dim_2), ctypes.c_void_p(tree.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(medians_dim_1), ctypes.c_void_p(medians.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(leaf_size), ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(dists_dim_1), ctypes.c_void_p(dists.ctypes.data), ctypes.byref(randomness), ctypes.byref(checks_present), ctypes.byref(checks), ctypes.byref(found_present), ctypes.byref(found), ctypes.byref(pt_ss_present), ctypes.byref(pt_ss), ctypes.byref(d_root_present), ctypes.byref(d_root))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, dists, (checks.value if checks_present else None), (found.value if found_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine FIX_ORDER
    
    def fix_order(self, points, sq_sums, radii, medians, order, copy=None):
        '''! Reorganize a built tree so that it is packed in order in memory.'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "medians"
        if ((not issubclass(type(medians), numpy.ndarray)) or
            (not numpy.asarray(medians).flags.f_contiguous) or
            (not (medians.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'medians' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            medians = numpy.asarray(medians, dtype=ctypes.c_float, order='F')
        medians_dim_1 = ctypes.c_long(medians.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "copy"
        copy_present = ctypes.c_bool(True)
        if (copy is None):
            copy_present = ctypes.c_bool(False)
            copy = ctypes.c_int()
        else:
            copy = ctypes.c_int(copy)
        if (type(copy) is not ctypes.c_int): copy = ctypes.c_int(copy)
    
        # Call C-accessible Fortran wrapper.
        clib.c_fix_order(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(medians_dim_1), ctypes.c_void_p(medians.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(copy_present), ctypes.byref(copy))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return points, sq_sums, radii, medians, order

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine BINCOUNT
    
    def bincount(self, indices, usage):
        '''! Increment the counts for the number of times various indices are referenced.
!   Example:
!     usage = [0, 0, 0, 0, 0]  ! counters for ball tree over 5 points
!     indices = [1, 4, 2, 1]   ! indices of points to increment
!     CALL BINCOUNT(indices, usage)
!     usage = [2, 1, 0, 0, 1]  ! updated counters for usage over 5 points'''
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "usage"
        if ((not issubclass(type(usage), numpy.ndarray)) or
            (not numpy.asarray(usage).flags.f_contiguous) or
            (not (usage.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'usage' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            usage = numpy.asarray(usage, dtype=ctypes.c_long, order='F')
        usage_dim_1 = ctypes.c_long(usage.shape[0])
    
        # Call C-accessible Fortran wrapper.
        clib.c_bincount(ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(usage_dim_1), ctypes.c_void_p(usage.ctypes.data))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return usage

ball_tree = ball_tree()




# This file is for testing the logic of partial aggregation, how the evaluation
#  is done, and most importantly how the gradient is calculated.

# Define a function for mapping integers to letters.
# The sequence created should look like:
#   0 -> A
#   1 -> B
#   ...
#   26 -> AA
#   27 -> AB
#   ..
#   52 -> BA
#   53 -> BB
def to_letter(number):
    letters = ""
    while number >= 0:
        remainder = number % 26
        letters = chr(ord("A") + remainder) + letters
        number = number // 26 - 1
    return letters

# Define a simple class that we can use to track what unique values
#  exist in this computation graph, the parents, children, and
#  the gradients. Includes methods for operations such as:
# 
#   - addition with another Value object
#   - division by another Value object
#   - setting a Number as the current contents
#   - string formatting of the computation defined by Value
#   - initializing with a default and unique letter name,
#     the letter is used to reference each instance when
#     displaying analytically computed values
#   - gradient evaluation with respect to a subset of children
# 
class Value:
    _content = None
    stop = False
    instances = []
    def __init__(self, content=None, letter=None):
        self.parents = []
        self.children = []
        self.operation = None
        if (letter is None):
            letter = to_letter(len(Value.instances))
        self.letter = letter
        self.content = content
        # Add this to list of global instances.
        Value.instances.append(self)

    # Define a "content" property that provides 
    @property
    def content(self):
        return (self if (self._content is None) else self._content)
    @content.setter
    def content(self, content):
        if (content is None):
            pass
        elif (not isinstance(content, (int, float))):
            raise(ValueError(f"{type(self).__name__}.content can only be assigned as an 'int' or 'float', but got '{type(content).__name__}'."))
        self._content = content

    # Return either the letter for self (if undefined) or the literal value.
    def __repr__(self):
        if (self._content is None):
            return self.letter
        else:
            return repr(self._content)

    # Return a nicely formatted string representation of how this Value is computed.
    def __str__(self):
        # If this is a "stop" variable, do not traverse through parents.
        if (self.stop):
            return repr(self)
        # Otherwise, get the full definition of this variable.
        if (self.operation == "+"):
            left = self.parents[0]
            right = self.parents[1]
            if ((left.content != 0) and (right.content != 0)):
                left = str(left)
                right = str(right)
                if ((left[0] == "(") and (left[-1] == ")")):
                    left = left[1:-1]
                if ((right[0] == "(") and (right[-1] == ")")):
                    right = right[1:-1]
                result = f"{left} {self.operation} {right}"
                if (len(self.children) > 0):
                    result = "(" + result + ")"
                return result
            elif (left.content != 0):
                return str(left)
            elif (right.content != 0):
                return str(right)
            else:
                return ""
        elif (self.operation == "/"):
            left = self.parents[0]
            right = self.parents[1]
            if (left.content == 0):
                return "0"
            elif (right.content == 0):
                return "NaN"
            elif (right.content == 1):
                return str(left)
            else:
                result = f"{left} {self.operation} {right}"
                if (len(self.children) > 0):
                    result = "(" + result + ")"
                return result
        else:
            return repr(self)

    # Add another Value to self.
    def __add__(self, other):
        if isinstance(other, (int, float)):
            other = Value(other)
        result = Value()
        result.parents = [self, other]
        result.operation = "+"
        other.children.append(result)
        self.children.append(result)
        return result
    def __radd__(self, other):
        return self.__add__(other)

    # Divide self by another Value.
    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            other = Value(other)
        result = Value()
        result.parents = [self, other]
        result.operation = "/"
        self.children.append(result)
        other.children.append(result)
        return result

    # Evaluate the experession defined by this Value.
    def __call__(self):
        if (self.operation == "+"):
            return self.parents[0].content + self.parents[1].content
        elif (self.operation == "/"):
            return self.parents[0].content / self.parents[1].content
        else:
            return self.content

    # Compute a new Value that is the gradient defined in terms of children.
    # Example of expected behavior is as follows:
    #   D = (A + B + C) / 3
    #   E = (A + B) / 2
    #   F = A
    # 
    #   C.gradient = D / 3
    #   B.gradient = D / 3 + E / 2
    #   A.gradeitn = D / 3 + E / 2 + F
    # 
    def gradient(self):
        # If there are no children, this is a leaf. Return a new Value with the
        #  same letter representation (returning 'self' could have side effects).
        if len(self.children) == 0:
            return Value(letter=self.letter)
        # Otherwise, this is an internal node that has children.
        else:
            # Initialize the gradient to 0.
            gradient = Value(0)
            for child in self.children:
                if child.operation == "+":
                    # Sums just add to the gradient for the parent.
                    gradient += child.gradient()
                elif child.operation == "/":
                    numerator, denominator = child.parents
                    if (numerator == self):
                        # The derivative of (a/b) with respect to (a) is (1/b).
                        gradient += child.gradient() / denominator
                    else:
                        # The derivative of (a/b) with respect to (b) is (-a/(b^2)).
                        gradient -= (child.gradient() * numerator) / (denominator * denominator)
                else:
                    raise ValueError(f"Unrecognized operation '{child.operation}', cannot compute gradient.")
            return gradient



# Declare the number of elements being aggregated.
n = 3

# Create a holder for the raw values.
inputs = [Value() for i in range(n)]

# Perform the partial aggregation calculation using
#   a range that iterates from [1, ..., n].
partials = [sum(inputs[-i:]) / i for i in range(1, n+1)]

# Print the computation graph for each of the partials.
print("Partials:")
for p in partials:
    value = str(p)
    p.stop = True
    print(p, "=", value)

print()
print("Gradients:")
for i in inputs:
    print(i.letter, "=", i.gradient())

